@@ -12,7 +12,7 @@ import functools, html.parser, itertools, json, math, re, threading, time, uuid
 import miniirc, requests, traceback  # type: ignore
 
 
-ver = (0, 0, 2)
+ver = (0, 0, 4)
 __version__ = '.'.join(map(str, ver))
 
 
@@ -99,9 +99,9 @@ class _TagManager:
     def set_colours(self, fg: Optional[str] = None,
                     bg: Optional[str] = None) -> None:
         if fg is not None:
-            self.fg = fg
+            self.fg = fg or None
         if bg is not None:
-            self.bg = bg
+            self.bg = bg or None
 
         if not self.fg and not self.bg:
             if 'span' in self.tags:
@@ -388,24 +388,44 @@ class Matrix(miniirc.IRC):
         self.__session.headers['Authorization'] = f'Bearer {token}'
         if auto_connect:
             self.connect()
+        else:
+            self._update_baseurl()
 
-    def _url_for(self, endpoint: str) -> str:
-        return f'https://{self.ip}:{self.port}/_matrix/client/v3/{endpoint}'
+    def _update_baseurl(self) -> None:
+        hostname = f'{self.ip}:{self.port}'
+        if (not self.ssl and self.ssl is not None and
+                self.ip in ('localhost', '127.0.0.1', '::1', '0::1')):
+            # Non-SSL localhost connections are probably to
+            # https://github.com/matrix-org/pantalaimon which doesn't support
+            # the "v3" URLs yet.
+            protocol = 'http'
+            api_version = 'r0'
+        else:
+            protocol = 'https'
+            api_version = 'v3'
+
+            # Use a shorter hostname if possible
+            if self.port == 443:
+                hostname = self.ip
+
+        matrix_url = f'{protocol}://{hostname}/_matrix'
+        self._baseurl = f'{matrix_url}/client/{api_version}'
+        self._media_baseurl = f'{matrix_url}/media/{api_version}'
 
     def __get(self, endpoint: str, timeout: int = 5, /,
               **params: Optional[str | int]) -> Any:
         self.debug('GET', endpoint, params)
-        return self.__session.get(self._url_for(endpoint), params=params,
+        return self.__session.get(f'{self._baseurl}/{endpoint}', params=params,
                                   timeout=timeout).json()
 
     def __post(self, endpoint: str, /, **params: Any) -> Any:
         self.debug('POST', endpoint, params)
-        return self.__session.post(self._url_for(endpoint), json=params,
+        return self.__session.post(f'{self._baseurl}/{endpoint}', json=params,
                                    timeout=5).json()
 
     def __put(self, endpoint: str, /, **params: Any) -> Any:
         self.debug('PUT', endpoint, params)
-        return self.__session.put(self._url_for(endpoint), json=params,
+        return self.__session.put(f'{self._baseurl}/{endpoint}', json=params,
                                   timeout=5).json()
 
     def _get_room_url_no_cache(self, room_id: str) -> str:
@@ -427,6 +447,7 @@ class Matrix(miniirc.IRC):
         if self.connected is not None:
             return
         with self._send_lock:
+            self._update_baseurl()
             self.active_caps = self.ircv3_caps & {
                 'account-tag', 'echo-message', 'message-tags',
             }
@@ -439,16 +460,7 @@ class Matrix(miniirc.IRC):
 
     def _main(self) -> None:
         try:
-            if miniirc.ver >= (2, 0, 0):
-                self.handle_msg(miniirc.IRCMessage('001', ('', '', ''), {}, [
-                    self.current_nick,
-                    f'Welcome to Matrix {self.current_nick}'
-                ]))
-            else:
-                self._handle('001', ('001', '001', '001'), {}, [
-                    self.current_nick,
-                    f':Welcome to Matrix {self.current_nick}'
-                ])
+            self.__numeric('001', f'Welcome to Matrix {self.current_nick}')
 
             next_batch: Optional[str] = None
             while self.connected:
@@ -510,7 +522,7 @@ class Matrix(miniirc.IRC):
     def quote(self, *msg: str, force: Optional[bool] = None,
               tags: Optional[dict[Any, Any]] = None) -> None:
         cmd, _, tags2, args = miniirc.ircv3_message_parser(' '.join(msg))
-        if args and args[-1].startswith(':'):
+        if miniirc.ver[0] < 2 and args and args[-1].startswith(':'):
             args[-1] = args[-1][1:]
         self.send(cmd, *args, force=force, tags=tags or tags2)
 
@@ -557,6 +569,9 @@ class Matrix(miniirc.IRC):
             self.debug(self.__post(f'join/{_url_quote(args[0])}'))
         elif cmd == 'PART' and len(args) == 1:
             self.debug(self.__post(f'{self._get_room_url(args[0])}/leave'))
+        elif self.connected:
+            self.debug('Unknown command:', cmd)
+            self.__numeric('421', cmd, 'Unknown command')
 
     def __send_tagmsg(self, channel: str, tags: dict[Any, Any]) -> None:
         if tags.get('+draft/react') and tags.get('+draft/reply'):
@@ -583,6 +598,11 @@ class Matrix(miniirc.IRC):
             self.handle_msg(miniirc.IRCMessage(
                 command, (sender, sender, sender), tags, args
             ))
+
+        def __numeric(self, numeric: str, *args: str) -> None:
+            self.handle_msg(miniirc.IRCMessage(
+                numeric, ('', '', ''), {}, [self.current_nick, *args]
+            ))
     else:
         def __irc_msg(self, event: _Event, command: str, args: list[str],
                       tags: Optional[dict[str, str]] = None, *,
@@ -597,6 +617,11 @@ class Matrix(miniirc.IRC):
 
             self._handle(command, (sender, sender, sender), tags, args)
 
+        def __numeric(self, numeric: str, *args: str) -> None:
+            raw_args = [self.current_nick, *args]
+            raw_args[-1] = ':' + raw_args[-1]
+            self._handle(numeric, (numeric, numeric, numeric), {}, raw_args)
+
     @_register_event('m.room.message')
     def _message_event(self, room_id: str, event: _Event) -> None:
         if ('echo-message' not in self.active_caps and
@@ -610,8 +635,7 @@ class Matrix(miniirc.IRC):
         if 'url' in content:
             msg = content.url[str]
             if msg.startswith('mxc://'):
-                msg = (f'https://{self.ip}:{self.port}/_matrix/media/v3/'
-                       f'download/{msg[6:]}')
+                msg = f'{self._media_baseurl}/download/{msg[6:]}'
         else:
             msg, html_parsed_ok = _matrix_html_to_irc(content)
 
@@ -673,6 +697,19 @@ class Matrix(miniirc.IRC):
             '+draft/reply': relates_to.event_id[str]
         })
 
+    @_register_event('im.vector.modular.widgets')
+    def _widget_event(self, room_id: str, event: _Event) -> None:
+        if event.content.type == 'jitsi':
+            data = event.content.data
+            msg = (f'\x01ACTION started a video conference: https://'
+                   f'{data.domain[str]}/{data.conferenceId[str]}\x01')
+        elif event.unsigned.prev_content.type == 'jitsi':
+            msg = '\x01ACTION ended a video conference\x01'
+        else:
+            return
+
+        self.__irc_msg(event, 'PRIVMSG', [room_id, msg], {})
+
     # Helpers
     @classmethod
     def _login(cls, homeserver: str, username: str, password: str) -> str:
@@ -697,7 +734,7 @@ class Matrix(miniirc.IRC):
     def _logout_all(cls, homeserver: str, username: str,
                     password: str) -> None:
         """ Logs out everywhere / voids all tokens for the user. """
-        token = cls.login(homeserver, username, password)
+        token = cls._login(homeserver, username, password)
         matrix = Matrix(homeserver, token=token, auto_connect=False)
         res = matrix.__post('logout/all')
         assert 'error' not in res, res
